﻿''' <summary>
''' The main calculator stack
''' </summary>
''' <remarks>Stack contains 4 registers (x, y, z, and t)</remarks>
Public Class CStack

    Private Const STACK_SIZE As Integer = 4

    Private reg(STACK_SIZE - 1) As Register

    ''' <summary>
    ''' The default constructor
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub New()
        For i As Integer = 0 To STACK_SIZE - 1
            reg(i) = New Register
        Next
    End Sub

    ''' <summary>
    ''' Build a new stack with the given Big Integer size
    ''' </summary>
    ''' <param name="size">The size in bits of the Big Integer</param>
    ''' <remarks></remarks>
    Public Sub New(size As Integer, mode As BigInt.ArithMode)
        For i As Integer = 0 To STACK_SIZE - 1
            reg(i) = New Register(size, mode)
        Next
    End Sub

    ''' <summary>
    ''' Peek at the value at the "bottom" of the stack
    ''' </summary>
    ''' <returns>The register at the bottom of the stack</returns>
    ''' <remarks>Does not lift or drop the stack</remarks>
    Public Function Peek() As Register
        Return reg(0)
    End Function

    ''' <summary>
    ''' Get a Register value off the stack
    ''' </summary>
    ''' <returns>The register value from the bottom of the stack</returns>
    ''' <remarks>Stack drops, but the top of the stack is duplicated</remarks>
    Public Function Pop() As Register
        Dim prev As Register

        prev = reg(0)
        ' drop the stack
        ' Note: the top of the stack is preserved
        For i As Integer = 0 To 2
            reg(i) = reg(i + 1)
        Next
        ' Make "deep copy" so 2 and 3 aren't pointing to the same
        ' register (since register is a reference type).
        reg(3) = reg(3).Copy
        Return prev
    End Function

    ''' <summary>
    ''' Push a value onto the stack
    ''' </summary>
    ''' <param name="val">The register to push onto the stack</param>
    ''' <remarks>Stack lifts, but the top of the stack is lost</remarks>
    Public Sub Push(val As Register)
        ' push a value on the stack
        ' Note: the top of the stack is lost
        For i As Integer = 2 To 0 Step -1
            reg(i + 1) = reg(i)
        Next
        reg(0) = val
    End Sub

    ''' <summary>
    ''' The X register
    ''' </summary>
    ''' <value>The new X register</value>
    ''' <returns>The current X register</returns>
    ''' <remarks></remarks>
    Public Property X As Register
        Get
            Return reg(0)
        End Get
        Set(value As Register)
            reg(0) = value
        End Set
    End Property

    ''' <summary>
    ''' The Y register
    ''' </summary>
    ''' <value>The new Y register</value>
    ''' <returns>The current Y register</returns>
    ''' <remarks></remarks>
    Public Property Y As Register
        Get
            Return reg(1)
        End Get
        Set(value As Register)
            reg(1) = value
        End Set
    End Property

    ''' <summary>
    ''' The Z register
    ''' </summary>
    ''' <value>The new Z register</value>
    ''' <returns>The current Z register</returns>
    ''' <remarks></remarks>
    Public Property Z As Register
        Get
            Return reg(2)
        End Get
        Set(value As Register)
            reg(2) = value
        End Set
    End Property

    ''' <summary>
    ''' The T register
    ''' </summary>
    ''' <value>The new T register</value>
    ''' <returns>The current T register</returns>
    ''' <remarks></remarks>
    Public Property T As Register
        Get
            Return reg(3)
        End Get
        Set(value As Register)
            reg(3) = value
        End Set
    End Property

    ''' <summary>
    ''' Copy the stack to an ordinary array
    ''' </summary>
    ''' <returns>An array of Register values</returns>
    ''' <remarks>A helper function to allow for easy XML serialization</remarks>
    Public Function ToArray() As Register()
        Dim ans(STACK_SIZE - 1) As Register
        For i As Integer = 0 To STACK_SIZE - 1
            ans(3 - i) = reg(i)
        Next
        Return ans
    End Function

End Class


