﻿Imports System.IO
Imports System.Xml.Serialization
Imports System.Text
Imports System.Xml

''' <summary>
''' The internal state of the calculator.  These setting are what get
''' stored while the calculator is turned off.
''' </summary>
''' <remarks></remarks>
Public Class CalcState

    Private Const MIN_FLOAT As Integer = 0
    Private Const MAX_FLOAT As Integer = 9
    Private Const KeyDp As Integer = &H48
    Private Const MIN_WORD As Integer = 1
    Private Const NUM_FLAGS As Integer = 6

    ''' <summary>
    ''' The major operating modes of the calculator
    ''' </summary>
    ''' <remarks></remarks>
    Public Enum CalcOpMode As Integer
        Float
        Hex
        Dec
        Oct
        Bin
    End Enum


    ''' <summary>
    ''' The calculator flags
    ''' </summary>
    ''' <remarks></remarks>
    Public Enum CalcFlag As Integer
        User0
        User1
        User2
        LeadingZero
        Carry
        Overflow
    End Enum

    ' In the real calculator the number of registers and program memory
    ' "compete" for the same 203 bytes.  So the max you can have are:
    ' 1) 406 registers (at 4-bit word size) and 0 program memory
    ' 2) 0 registers and 203 program memory
    '
    ' However, in this version, there is no competition and the numbers
    ' are static.  You can choose any numbers you like in the config file.
    ' The default is 32 registers and 203 lines of memory.


    Private pfvSaveOnExit As Boolean
    Private pfvWordSize As Integer
    Private pfvOpMode As CalcState.CalcOpMode
    Private pfvArithMode As BigInt.ArithMode
    Private pfvFloatPrecision As Integer
    Private pfvFlags() As Boolean
    Private pfvReg() As Register
    Private pfvRegIndex As Register
    Private pfvStack As CStack
    Private pfvRegLastX As Register
    Private pfvPrgmPosition As Integer
    Private pfvPrgmMemory As List(Of String)
    Private pfvPrgmRetStack As Stack(Of Integer)
    Private pfvPrgmRunning As Boolean

    Private numRegs As Integer
    Private syncConv As Boolean
    Private numLines As Integer

    ''' <summary>
    ''' Default constructor
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub New()

        numRegs = My.Settings.NumRegisters
        syncConv = My.Settings.SyncConversions
        numLines = My.Settings.PrgmMemoryLines

        pfvFlags = New Boolean(NUM_FLAGS - 1) {}
        pfvReg = New Register(numRegs - 1) {}

        pfvSaveOnExit = True
        pfvWordSize = 16
        pfvOpMode = CalcState.CalcOpMode.Float
        pfvArithMode = BigInt.ArithMode.TwosComp
        pfvFloatPrecision = 3
        For i As Integer = 0 To NUM_FLAGS - 1
            pfvFlags(i) = False
        Next
        pfvFlags(3) = True ' leading zeros
        For i As Integer = 0 To numRegs - 1
            pfvReg(i) = New Register(pfvWordSize, pfvArithMode)
        Next
        pfvRegIndex = New Register(64, pfvArithMode)        ' RegIndex has a fixed size
        pfvStack = New CStack(pfvWordSize, pfvArithMode)
        pfvRegLastX = New Register(pfvWordSize, pfvArithMode)
        pfvPrgmPosition = 0
        pfvPrgmMemory = New List(Of String)
        pfvPrgmRetStack = New Stack(Of Integer)
        pfvPrgmRunning = False
    End Sub

    ''' <summary>
    ''' Save the configuration on Exit
    ''' </summary>
    ''' <value>The new exit setting</value>
    ''' <returns>The existing exit setting</returns>
    ''' <remarks></remarks>
    Public Property SaveOnExit As Boolean
        Get
            Return pfvSaveOnExit
        End Get
        Set(value As Boolean)
            pfvSaveOnExit = value
        End Set
    End Property

    ''' <summary>
    ''' The current operating mode of the calculator
    ''' </summary>
    ''' <value>Set the operating mode</value>
    ''' <returns>Return the current operating mode</returns>
    ''' <remarks></remarks>
    Public Property OpMode As CalcState.CalcOpMode
        Get
            Return pfvOpMode
        End Get
        Set(value As CalcState.CalcOpMode)
            pfvOpMode = value
        End Set
    End Property

    ''' <summary>
    ''' The Arithmetic mode (the complement mode)
    ''' </summary>
    ''' <value>Set the Arithmetic mode</value>
    ''' <returns>Return the current Arithmetic mode</returns>
    ''' <remarks></remarks>
    Public Property ArithMode As BigInt.ArithMode
        Get
            Return pfvArithMode
        End Get
        Set(value As BigInt.ArithMode)
            If pfvArithMode <> value Then
                ' change the Arithmetic Mode of everything
                ReArithAll(value)
            End If
            pfvArithMode = value
        End Set
    End Property

    ''' <summary>
    ''' The number of decimal points for the display
    ''' </summary>
    ''' <value>Set the decimal points</value>
    ''' <returns>The current decimal points</returns>
    ''' <remarks>Affects display only (Ingored in integer mode)</remarks>
    Public Property FloatPrecision As Integer
        Get
            Return pfvFloatPrecision
        End Get
        Set(value As Integer)
            If value >= MIN_FLOAT And (value <= MAX_FLOAT Or value = KeyDp) Then
                pfvFloatPrecision = value
            End If
        End Set
    End Property

    ''' <summary>
    ''' The size of the integers
    ''' </summary>
    ''' <value>Set the word size</value>
    ''' <returns>The current word size</returns>
    ''' <remarks>Ingored in float mode</remarks>
    Public Property WordSize As Integer
        Get
            Return pfvWordSize
        End Get
        Set(value As Integer)
            If value >= MIN_WORD Then
                If pfvWordSize <> value Then
                    ' change the size of everything
                    ReSizeAll(value)
                End If
                pfvWordSize = value
            End If
        End Set
    End Property

    ''' <summary>
    ''' The 6 user and systems flags
    ''' </summary>
    ''' <value>Set the flags</value>
    ''' <returns>Return the flags</returns>
    ''' <remarks>6 flags over texas :)</remarks>
    Public Property Flags As Boolean()
        Get
            Return pfvFlags
        End Get
        Set(value As Boolean())
            pfvFlags = value
        End Set
    End Property

    ''' <summary>
    ''' The storage registers
    ''' </summary>
    ''' <value>The new storage registers</value>
    ''' <returns>Return the storage registers</returns>
    ''' <remarks>Only the first 32 are directly accessible</remarks>
    Public Property Reg As Register()
        Get
            Return pfvReg
        End Get
        Set(value As Register())
            pfvReg = value
        End Set
    End Property

    ''' <summary>
    ''' The index register
    ''' </summary>
    ''' <value>Set the value for the I register</value>
    ''' <returns>Return the value for the I register</returns>
    ''' <remarks>
    ''' Is not altered during integer and float conversions.
    ''' </remarks>
    Public Property RegIndex As Register
        Get
            Return pfvRegIndex
        End Get
        Set(value As Register)
            pfvRegIndex = value
        End Set
    End Property

    ''' <summary>
    ''' The Last X register
    ''' </summary>
    ''' <value>Set the Last X value</value>
    ''' <returns>Return the Last X value</returns>
    ''' <remarks></remarks>
    Public Property RegLastX As Register
        Get
            Return pfvRegLastX
        End Get
        Set(value As Register)
            pfvRegLastX = value
        End Set
    End Property

    ''' <summary>
    ''' The location (the line) where the program is currently executing
    ''' </summary>
    ''' <value>Set the program location</value>
    ''' <returns>Return the program location</returns>
    ''' <remarks></remarks>
    Public Property PrgmPosition As Integer
        Get
            Return pfvPrgmPosition
        End Get
        Set(value As Integer)
            If value >= 0 And value <= numLines Then
                pfvPrgmPosition = value
            End If
        End Set
    End Property

    ''' <summary>
    ''' Program memory
    ''' </summary>
    ''' <value>Set the program memory</value>
    ''' <returns>Return the program memory</returns>
    ''' <remarks></remarks>
    Public Property PrgmMemory As List(Of String)
        Get
            Return pfvPrgmMemory
        End Get
        Set(value As List(Of String))
            pfvPrgmMemory = value
        End Set
    End Property

    ''' <summary>
    ''' Program Return Stack
    ''' </summary>
    ''' <value>Set the Program Return stack</value>
    ''' <returns>Get the Program Return stack</returns>
    ''' <remarks>Is serialized thru an array</remarks>
    Public Property PrgmRetStack As Stack(Of Integer)
        Get
            Return pfvPrgmRetStack
        End Get
        Set(value As Stack(Of Integer))
            pfvPrgmRetStack = value
        End Set
    End Property

    ''' <summary>
    ''' The Calculator Stack (x, y, z, and t registers)
    ''' </summary>
    ''' <value>Set the Stack</value>
    ''' <returns>Get the Stack</returns>
    ''' <remarks>Is serialized thru an array</remarks>
    Public Property Stack As CStack
        Get
            Return pfvStack
        End Get
        Set(value As CStack)
            pfvStack = value
        End Set
    End Property

    ''' <summary>
    ''' Is the calculator currently running a program
    ''' </summary>
    ''' <value>The new running value</value>
    ''' <returns>the current running value</returns>
    ''' <remarks>set to false to terminate a running program</remarks>
    Public Property PrgmRunning As Boolean
        Get
            Return pfvPrgmRunning
        End Get
        Set(value As Boolean)
            pfvPrgmRunning = value
        End Set
    End Property

    ''' <summary>
    ''' Save the CalcState as an XML string
    ''' </summary>
    ''' <returns>An XML string suitable for saving</returns>
    ''' <remarks>No error catching is done here</remarks>
    Public Function Serialize() As String
        Dim sw As New StringWriter
        Dim xtw As New XmlTextWriter(sw)
        Dim xd As New XmlDocument
        Dim xn, xn1, xn2 As XmlNode
        Dim num As Integer

        ' create the preamble
        xd.AppendChild(xd.CreateXmlDeclaration("1.0", "UTF-8", "yes"))

        ' create the root element
        xn = xd.AppendChild(xd.CreateElement("CalcState"))
        xn.Attributes.Append(xd.CreateAttribute("saved")).InnerText = Now.ToString("s")

        xn.AppendChild(xd.CreateComment("WRPN CalcState v" & My.Application.Info.Version.Major & "." & My.Application.Info.Version.Minor))
        xn.AppendChild(xd.CreateElement("SaveOnExit")).InnerText = pfvSaveOnExit.ToString.ToLower
        xn.AppendChild(xd.CreateElement("WordSize")).InnerText = pfvWordSize.ToString
        xn.AppendChild(xd.CreateElement("OpMode")).InnerText = pfvOpMode.ToString
        xn.AppendChild(xd.CreateElement("ArithMode")).InnerText = pfvArithMode.ToString
        xn.AppendChild(xd.CreateElement("FloatPrecision")).InnerText = pfvFloatPrecision.ToString
        Dim FlagName() As String = {"User0", "User1", "User2", "LeadingZero", "Carry", "Overflow"}
        xn1 = xn.AppendChild(xd.CreateElement("Flags"))
        For i As Integer = 0 To NUM_FLAGS - 1
            xn2 = xn1.AppendChild(xd.CreateElement("Flag"))
            xn2.InnerText = pfvFlags(i).ToString.ToLower
            xn2.Attributes.Append(xd.CreateAttribute("name")).InnerText = FlagName(i)
        Next
        Dim RegName() As String = {"0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "A", "B", "C", "D", "E", "F",
                                   ".0", ".1", ".2", ".3", ".4", ".5", ".6", ".7", ".8", ".9", ".A", ".B", ".C", ".D", ".E", ".F"}

        xn1 = xn.AppendChild(xd.CreateElement("Regs"))
        For i As Integer = 0 To My.Settings.NumRegisters - 1
            xn2 = xn1.AppendChild(xd.CreateElement("Reg"))
            If i < RegName.Length Then
                xn2.Attributes.Append(xd.CreateAttribute("name")).InnerText = RegName(i)
            Else
                xn2.Attributes.Append(xd.CreateAttribute("name")).InnerText = "Reg" & i.ToString
            End If
            xn2.AppendChild(xd.CreateElement("FVal")).InnerText = pfvReg(i).FVal.ToString(System.Globalization.CultureInfo.InvariantCulture)
            xn2.AppendChild(xd.CreateElement("BiVal")).InnerText = pfvReg(i).BiVal.ToStringHex
        Next
        xn1 = xn.AppendChild(xd.CreateElement("RegIndex"))
        xn1.AppendChild(xd.CreateElement("FVal")).InnerText = pfvRegIndex.FVal.ToString(System.Globalization.CultureInfo.InvariantCulture)
        xn1.AppendChild(xd.CreateElement("BiVal")).InnerText = pfvRegIndex.BiVal.ToStringHex

        num = 0
        Dim StackName() As String = {"T", "Z", "Y", "X"}
        xn1 = xn.AppendChild(xd.CreateElement("Stacks"))
        For Each r As Register In pfvStack.ToArray
            xn2 = xn1.AppendChild(xd.CreateElement("Stack"))
            xn2.Attributes.Append(xd.CreateAttribute("name")).InnerText = StackName(num)
            xn2.AppendChild(xd.CreateElement("FVal")).InnerText = r.FVal.ToString(System.Globalization.CultureInfo.InvariantCulture)
            xn2.AppendChild(xd.CreateElement("BiVal")).InnerText = r.BiVal.ToStringHex
            num += 1
        Next

        xn1 = xn.AppendChild(xd.CreateElement("RegLastX"))
        xn1.AppendChild(xd.CreateElement("FVal")).InnerText = pfvRegLastX.FVal.ToString(System.Globalization.CultureInfo.InvariantCulture)
        xn1.AppendChild(xd.CreateElement("BiVal")).InnerText = pfvRegLastX.BiVal.ToStringHex

        xn.AppendChild(xd.CreateElement("PrgmPosition")).InnerText = pfvPrgmPosition.ToString

        ' start with 1, since that's what the display uses
        num = 1
        xn1 = xn.AppendChild(xd.CreateElement("PrgmMemory"))
        For Each line As String In pfvPrgmMemory
            xn2 = xn1.AppendChild(xd.CreateElement("Line"))
            xn2.Attributes.Append(xd.CreateAttribute("name")).InnerText = num.ToString("000")
            xn2.InnerText = line
            num += 1
        Next

        num = 0
        Dim temp2() As Integer = pfvPrgmRetStack.ToArray
        Array.Reverse(temp2)
        xn1 = xn.AppendChild(xd.CreateElement("PrgmRetStack"))
        For Each i As Integer In temp2
            xn2 = xn1.AppendChild(xd.CreateElement("Return"))
            xn2.Attributes.Append(xd.CreateAttribute("name")).InnerText = num.ToString
            xn2.InnerText = i.ToString
            num += 1
        Next

        xtw.Formatting = Formatting.Indented
        xd.WriteTo(xtw)
        Return sw.ToString
    End Function

    ''' <summary>
    ''' Convert a saved XML file into the CalcState
    ''' </summary>
    ''' <remarks>No error catching is done here</remarks>
    Public Sub Deserialize(state As String)
        Dim xd As New XmlDocument
        Dim nl As XmlNodeList
        Dim xn As XmlNode
        Dim num As Integer

        xd.LoadXml(state)

        pfvSaveOnExit = CBool(xd.SelectSingleNode("CalcState/SaveOnExit").InnerText)
        pfvWordSize = Integer.Parse(xd.SelectSingleNode("CalcState/WordSize").InnerText)
        pfvOpMode = CType([Enum].Parse(GetType(CalcOpMode), xd.SelectSingleNode("CalcState/OpMode").InnerText), CalcOpMode)
        pfvArithMode = CType([Enum].Parse(GetType(BigInt.ArithMode), xd.SelectSingleNode("CalcState/ArithMode").InnerText), BigInt.ArithMode)
        pfvFloatPrecision = Integer.Parse(xd.SelectSingleNode("CalcState/FloatPrecision").InnerText)

        num = 0
        For Each xflag As XmlNode In xd.SelectNodes("CalcState/Flags/Flag")
            pfvFlags(num) = CBool(xflag.InnerText)
            num += 1
        Next

        nl = xd.SelectNodes("CalcState/Regs/Reg")
        ' it's possible that somebody changed the configuration, so we can't count on NumRegisters
        Dim num_regs As Integer = Math.Min(nl.Count, My.Settings.NumRegisters)
        For i As Integer = 0 To num_regs - 1
            xn = nl.Item(i)
            pfvReg(i).FVal = Double.Parse(xn.ChildNodes(0).InnerText, System.Globalization.CultureInfo.InvariantCulture)
            pfvReg(i).BiVal = New BigInt(xn.ChildNodes(1).InnerText, pfvWordSize, pfvArithMode)
        Next

        xn = xd.SelectSingleNode("CalcState/RegIndex")
        pfvRegIndex.FVal = Double.Parse(xn.ChildNodes(0).InnerText, System.Globalization.CultureInfo.InvariantCulture)
        pfvRegIndex.BiVal = New BigInt("&H" & xn.ChildNodes(1).InnerText, 64, pfvArithMode)

        nl = xd.SelectNodes("CalcState/Stacks/Stack")
        Dim r As Register
        For i As Integer = 0 To nl.Count - 1
            xn = nl.Item(i)
            r = New Register
            r.FVal = Double.Parse(xn.ChildNodes(0).InnerText, System.Globalization.CultureInfo.InvariantCulture)
            r.BiVal = New BigInt("&H" & xn.ChildNodes(1).InnerText, pfvWordSize, pfvArithMode)
            pfvStack.Push(r)
        Next

        xn = xd.SelectSingleNode("CalcState/RegLastX")
        pfvRegLastX.FVal = Double.Parse(xn.ChildNodes(0).InnerText, System.Globalization.CultureInfo.InvariantCulture)
        pfvRegLastX.BiVal = New BigInt("&H" & xn.ChildNodes(1).InnerText, pfvWordSize, pfvArithMode)

        pfvPrgmPosition = Integer.Parse(xd.SelectSingleNode("CalcState/PrgmPosition").InnerText)

        nl = xd.SelectNodes("CalcState/PrgmMemory/Line")
        PrgmMemory.Clear()
        If nl IsNot Nothing Then
            For i As Integer = 0 To nl.Count - 1
                xn = nl.Item(i)
                pfvPrgmMemory.Add(xn.InnerText)
            Next
        End If

        nl = xd.SelectNodes("CalcState/PrgmRetStack/Return")
        PrgmRetStack.Clear()
        If nl IsNot Nothing Then
            For i As Integer = 0 To nl.Count - 1
                xn = nl.Item(i)
                pfvPrgmRetStack.Push(Integer.Parse(xn.InnerText))
            Next
        End If
    End Sub

    ''' <summary>
    ''' Resize all of the Big Integer values inside the registers
    ''' </summary>
    ''' <param name="size">The new bitsize for the Big Integer's</param>
    ''' <remarks>The Index register is NOT resized</remarks>
    Private Sub ReSizeAll(size As Integer)
        If My.Settings.SyncConversions Then
            For i As Integer = 0 To My.Settings.NumRegisters - 1
                pfvReg(i).BiVal.WordSize = size
            Next
            pfvRegLastX.BiVal.WordSize = size
            pfvStack.X.BiVal.WordSize = size
            pfvStack.Y.BiVal.WordSize = size
            pfvStack.Z.BiVal.WordSize = size
            pfvStack.T.BiVal.WordSize = size
        Else
            ' Using the rules on page 32 and 66 the storage registers
            ' are not resized.  Also, the stack does not preserve the
            ' sign bit on widening
            Dim tempArith As BigInt.ArithMode = pfvArithMode
            pfvRegLastX.BiVal.BIArithMode = BigInt.ArithMode.Unsigned
            pfvStack.X.BiVal.BIArithMode = BigInt.ArithMode.Unsigned
            pfvStack.Y.BiVal.BIArithMode = BigInt.ArithMode.Unsigned
            pfvStack.Z.BiVal.BIArithMode = BigInt.ArithMode.Unsigned
            pfvStack.T.BiVal.BIArithMode = BigInt.ArithMode.Unsigned

            pfvRegLastX.BiVal.WordSize = size
            pfvStack.X.BiVal.WordSize = size
            pfvStack.Y.BiVal.WordSize = size
            pfvStack.Z.BiVal.WordSize = size
            pfvStack.T.BiVal.WordSize = size

            pfvRegLastX.BiVal.BIArithMode = tempArith
            pfvStack.X.BiVal.BIArithMode = tempArith
            pfvStack.Y.BiVal.BIArithMode = tempArith
            pfvStack.Z.BiVal.BIArithMode = tempArith
            pfvStack.T.BiVal.BIArithMode = tempArith
        End If

    End Sub

    ''' <summary>
    ''' Reset the Arithmetic mode of all of the BigInt values inside the registers
    ''' </summary>
    ''' <param name="mode">The Arithmetic mode for the Big Integer's</param>
    ''' <remarks></remarks>
    Private Sub ReArithAll(mode As BigInt.ArithMode)
        For i As Integer = 0 To numRegs - 1
            pfvReg(i).BiVal.BIArithMode = mode
        Next
        pfvRegLastX.BiVal.BIArithMode = mode
        pfvRegIndex.BiVal.BIArithMode = mode
        pfvStack.X.BiVal.BIArithMode = mode
        pfvStack.Y.BiVal.BIArithMode = mode
        pfvStack.Z.BiVal.BIArithMode = mode
        pfvStack.T.BiVal.BIArithMode = mode
    End Sub

    ''' <summary>
    ''' Synchronize the Big Integer and Float values within a Register
    ''' </summary>
    ''' <remarks>SyncConversions is used to determine the level of
    ''' adherence to the real calculator</remarks>
    Public Sub SyncValues()

        If My.Settings.SyncConversions Then
            ' With SyncConversion set to true, we do NOT follow the behavior
            ' of the real calculator.  Instead, we synchronize the integer
            ' and float values when switching between modes.  Obviously, there
            ' will likely be some loss of precision when decimal values are
            ' truncated.
            If OpMode = CalcState.CalcOpMode.Float Then
                ' if we're current in the float mode, then we need to
                ' copy those values to the to the integer mode
                For i As Integer = 0 To numRegs - 1
                    pfvReg(i).BiVal = New BigInt(pfvReg(i).FVal, pfvWordSize, pfvArithMode)
                Next
                pfvRegLastX.BiVal = New BigInt(pfvRegLastX.FVal, pfvWordSize, pfvArithMode)
                pfvStack.X.BiVal = New BigInt(pfvStack.X.FVal, pfvWordSize, pfvArithMode)
                pfvStack.Y.BiVal = New BigInt(pfvStack.Y.FVal, pfvWordSize, pfvArithMode)
                pfvStack.Z.BiVal = New BigInt(pfvStack.Z.FVal, pfvWordSize, pfvArithMode)
                pfvStack.T.BiVal = New BigInt(pfvStack.T.FVal, pfvWordSize, pfvArithMode)
                pfvRegIndex.BiVal = New BigInt(pfvRegIndex.FVal, 64, pfvArithMode)
            Else
                ' copy the values to the float mode
                For i As Integer = 0 To numRegs - 1
                    pfvReg(i).FVal = pfvReg(i).BiVal.ToLong
                Next
                pfvRegLastX.FVal = pfvRegLastX.BiVal.ToLong
                pfvStack.X.FVal = pfvStack.X.BiVal.ToLong
                pfvStack.Y.FVal = pfvStack.Y.BiVal.ToLong
                pfvStack.Z.FVal = pfvStack.Z.BiVal.ToLong
                pfvStack.T.FVal = pfvStack.T.BiVal.ToLong
                pfvRegIndex.FVal = pfvRegIndex.BiVal.ToLong
            End If
        Else
            ' If no "Sync Conversion" then we use the rules from the real calculator (yuk!)
            If OpMode = CalcState.CalcOpMode.Float Then
                ' The rules for converting from Float mode to Integer mode
                ' are found at page 60
                pfvWordSize = 56

                Dim x, temp As Double
                Dim tempX, tempY As Long
                Dim neg As Boolean

                If pfvStack.X.FVal <> 0.0 Then
                    x = pfvStack.X.FVal
                    If x < 0 Then
                        neg = True
                        x = x * -1
                    End If
                    temp = x / Math.Pow(2, 32)
                    tempX = CLng(Math.Log(temp) / Math.Log(2))
                    tempY = CLng(x / Math.Pow(2, tempX))
                    If neg Then
                        tempY = tempY * -1
                    End If
                End If

                pfvStack.X.BiVal = New BigInt(tempX, pfvWordSize, pfvArithMode)
                pfvStack.Y.BiVal = New BigInt(tempY, pfvWordSize, pfvArithMode)
            Else
                ' The rules for converting from Integer mode to Float mode
                ' are found at page 57-58
                pfvStack.Y.FVal = 0
                pfvStack.Z.FVal = 0
                pfvStack.T.FVal = 0
                pfvRegLastX.FVal = 0

                ' set x = (y)(2^x)
                pfvStack.X.FVal = pfvStack.Y.BiVal.ToLong * Math.Pow(2, pfvStack.X.BiVal.ToLong)
            End If
        End If
    End Sub
End Class


